/***************************************************************************
                          fehler.h  -  description
                             -------------------
    begin                : Fri Jul 21 2000
    copyright            : (C) 2000 by Immi
    email                : cuyo@pcpool.mathematik.uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef FEHLER_H
#define FEHLER_H

#include <qstring.h>
#include "inkompatibel.h"
#include "cuyointl.h"


/** Fehler, der gethrowt werden kann. Macht beim Erzeugen gleich ein
    bisschen Aufrumarbeiten: Blop::abbruchGleichzeitig() wird aufgerufen. */
struct Fehler {
  Fehler(const char * fmt, ...)
    /*__attribute__ ((format (printf, 2, 3)))*/;
    
  __String mText;
  /** True, wenn es schon einen Code gab, der gesagt hat, wo der Fehler
      passiert ist. (Damit so was nicht zweimal gesagt wird.) */
  bool mMitZeile;
};


/* PBEGIN_TRY und PEND_TRY() werden in scanner.ll und parser.yy verwendet.
   Sie fangen Fehler ab und geben sie aus, damit
   weitergeparst werden kann und mehrere Fehler gleichzeitig ausgegeben
   werden knnen. (P = Parse) */

#define PBEGIN_TRY try {
#define PEND_TRY(on_error) } catch (Fehler fe) {\
  fprintf(stderr, "%s:%d: %s\n", gDateiName.data(), gZeilenNr,\
                                 fe.mText.data());\
  gGabFehler = true;\
  { on_error; }\
}

/* TBEGIN_TRY und TEND_TRY() werden in cuyo.cpp verwendet.
   Der Fehler wird nach stderr und in das Cuyo-Fenster ausgegeben...
   bei anz_sp vielen Spielern. str wird auch noch ausgegeben. */

#define TBEGIN_TRY try {
#define TEND_TRY(anz_sp, str, on_error) } catch (Fehler fe) {\
  fprintf(stderr, "%s%s\n", str, fe.mText.data()); \
  for (int i = 0; i < anz_sp; i++) \
    /* zweites true = Kleine Schrift... */ \
    mSpielfeld[i]->setText(__String(str) + fe.mText, true, true); \
  { on_error; } \
}





/* Debug-Macro: Zhlt, wie oft es aufgerufen wird. Gibt nach schritte
   vielen Malen den neuen Wert und string aus. */
#define D_ZAEHLEN(schritte, string) do { \
  static int _zzz = 0;\
  if (_zzz++ % schritte == 0) fprintf(stderr, string" %d\n", _zzz);\
} while (0)



#define CASSERT(blub) do { \
  if (!(blub)) \
    throw Fehler(_("Internal error: \"%s\" in %s (%d)"), \
                  #blub, __FILE__, __LINE__); \
} while (0)


#endif
