/***************************************************************************
                          knoten.cpp  -  description
                             -------------------
    begin                : Mit Jul 12 22:54:51 MEST 2000
    copyright            : (C) 2000 by Immi
    email                : cuyo@pcpool.mathematik.uni-freiburg.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#include <cstdlib>
#include <cstdio>

#include "cuyointl.h"
#include "knoten.h"
#include "code.h"
#include "blop.h"
#include "leveldaten.h"



/*************************************************************************/
/* Was von vornerein so im Namespace rumliegt */


/* xxx_anz wird in blop.h definiert. */

char spezvar_namen[spezvar_anz][22] = {
  "file", "pos", "kind", "version", "qu",
  "out1", "out2"
};

int spezvar_default[spezvar_anz] = {
  0, 0, blopart_ausserhalb, 0, viertel_alle,
  spezvar_out_nichts, spezvar_out_nichts
};


/* Reihenfolge -1, -2, -3, ... */
char spezconst_namen[spezconst_anz][22] = {
  "turn", "connect", "falling", "size", "loc_x", "loc_y", "loc_p",
  "players"
};

int spezconst_default[spezconst_anz] = {
  0, 0, 0, 0, -1, -1, 0,
  0
};


int const_anz = (21+2);
char const_namen[27][22] = {
  /* Viertelstckchen */
  "Q_ALL",
  "Q_TL", "Q_TR", "Q_BL", "Q_BR",
  "Q_TL_TL", "Q_TR_TL", "Q_BL_TL", "Q_BR_TL",
  "Q_TL_TR", "Q_TR_TR", "Q_BL_TR", "Q_BR_TR",
  "Q_TL_BL", "Q_TR_BL", "Q_BL_BL", "Q_BR_BL",
  "Q_TL_BR", "Q_TR_BR", "Q_BL_BR", "Q_BR_BR",
  
  /* Verwandlungsziele */
  "nothing", "outside"
};

/* Eigentlich sollten hier die Konstanten aus bilddatei.h verwendet werden;
   das wre aber so viel Tipparbeit... */
int const_werte[26] = {
  /* Viertelstckchen */
  viertel_alle,
  0, 5, 10, 15,
  0, 1, 2, 3,
  4, 5, 6, 7,
  8, 9, 10, 11,
  12, 13, 14, 15,
  
  /* Verwandlungsziele */
  blopart_keins, blopart_ausserhalb
};



/*************************************************************************/
/* DefKnoten */


DefKnoten::DefKnoten(DefKnoten * vater):
  mVater(vater), mTiefe(vater ? vater->mTiefe + 1 : tiefe_global),
  /* Das ganze Variablen-Zeug vom Vater bernehmen...: */
  mVarNrBei(vater ? vater->mVarNrBei : 0),
  mBoolNrBei(vater ? vater->mBoolNrBei : -1),
  mDefaultWerte(vater ? vater->mDefaultWerte.copy() : QArray<int>())
{
  /* Im obersten Knoten die vordefinierten Namespace-Dinge einfgen. */
  if (mTiefe == tiefe_global) {
    speicherGlobaleVordefinierte();
  }
}

DefKnoten::~DefKnoten() {
  tKnotenMap::Iterator it;
  for(it = mKinder.begin(); it != mKinder.end(); ++it)
    delete it.data();
}




__String DefKnoten::toString() const {
  tKnotenMap::ConstIterator it;
  __String ret = "{\n";
  for(it = mKinder.begin(); it != mKinder.end(); ++it) {
    ret += it.key() + "=" + it.data()->toString() + "\n";
  }
  ret += "}\n";
  return ret;
}


void DefKnoten::fuegeEin(const __String & na, Knoten * wert) {
  if (mKinder.contains(na))
    throw Fehler(_("\"%s\" already defined."), na.data());
  mKinder[na] = wert;

  /* Wenn ein neuer Level geparst wird, will das ld evtl. wissen.
     (Falls der Benutzer neu gemachte Level ausprobieren will,
     die noch nicht in "level=..." stehen.) */
  if (mTiefe == tiefe_global && wert->type() == type_DefKnoten)
    ld->levelGefunden(na);
  
  /* Auf Level-Ebene evtl. noch ein paar Konstanten einfgen... */
  if (mTiefe == tiefe_level) {
    /* pics-Eintrag? Dann knnen ja jetzt die pics-Konstanten erzeugt
       werden. */

    /* Wenn beim Definieren der Konstanten ein mehrfach-definiert-Fehler
        auftritt, dann soll der nicht ausgegeben werden; die neue
	Definition wird einfach ignoriert. Das ist ntig bei irgend welchen
	Test-Leveln, wo fr mehrere Sorten einfach das gleiche verwendet
	wird. */
    try {
    
      /* Achtung, verwirrend: wert ist nicht der Wert der Konstante. */
      if (na == "pics")
	speicherPicsConst(wert);
      if (na == "greypic")
	speicherKnotenConst(wert, blopart_grau);
      if (na == "startpic")
	speicherKnotenConst(wert, blopart_gras);
      if (na == "emptypic")
	speicherKnotenConst(wert, blopart_keins);
    
    } catch (Fehler fe) {
    }
    
  }
}


/***** Methoden fr den Codespeicher *****/



/** Speichert alle vordefinierten Variablen in den
    Namespace, auer die pics-Konstanten. Wird vom Constructor
    des WurzelKnotens aufgerufen. */
void DefKnoten::speicherGlobaleVordefinierte() {

  /* Spezial-Variablen (z. B. file) */
  /* Das sollten die ersten Variablen sein, die erzeugt werden;
     sonst stimmen nachher die Nummern nicht. */
  CASSERT(mVarNrBei == 0);
  for (int i = 0; i < spezvar_anz; i++) {
    /* Nicht direkt mit speicherDefinition() erzeugen; sonst
       werden die Default-Werte nicht nochmal in mDefaultWerte
       gespeichert. */
    neueVarDefinition(spezvar_namen[i], spezvar_default[i]);
  }

  /* Spezial-Konstanten (d. h. Variablen, die nur gelesen werden
     werden knnen), z. B. verbindung */  
  for (int i = 0; i < spezconst_anz; i++)
    speicherDefinition(namespace_variable, spezconst_namen[i],
        new VarDefinition(spezconst_namen[i], spezconst_default[i],
                          vd_spezconst, -i-1)
      );
  
  /* Ganz normale feste Konstanten (z. B. nochange) */
  for (int i = 0; i < const_anz; i++)
    speicherDefinition(namespace_variable, const_namen[i],
        new VarDefinition(const_namen[i], const_werte[i],
                          vd_konstante)
      );

}



/** Speichert die Pics-Konstanten. (picsliste sollte der pics-Knoten sein.)
    Wird von fuegeEin(...) aufgerufen, wenn es die pics bekommt. */
void DefKnoten::speicherPicsConst(Knoten * picsliste) {
  /* pics-Konstanten. (Eigentlich knnten wir gleich eine Fehlermeldung
     ausspucken, wenn der Typ von picsliste falsch ist, aber die anderen
     entsprechenden Typ-Fehlermeldungen kommen auch erst beim Starten des
     Levels. Also hier erst mal ruhig sein. */
  if (picsliste->type() == type_ListenKnoten) {
    ListenKnoten * lki = (ListenKnoten*) picsliste;
    for (int i = 0; i < lki->getLaenge(); i++) {
      /* picsEndungWeg() steht in code.* */
      __String varna = picsEndungWeg(lki->getWort(i));
      speicherDefinition(namespace_variable, varna,
          new VarDefinition(varna, i, vd_konstante)
        );
     }
  }
}


/** Speichert eine Konstante mit dem Namen, der in nameKnoten steht und
    dem angegebenen Wert. nameKnoten ist hoffentlich ein ListenKnoten
    mit genau einem Eintrag. Wird von fuegeEin() aufgerufen, um die
    Gras-, die Grau- und die nix-Konstante abzuspeichern, wenn es die
    bekommt. */
void DefKnoten::speicherKnotenConst(Knoten * nameKnoten, int wert) {
  /* Keine Fehlermeldung bei falschem Typ; siehe speicherPicsConst() */
  if (nameKnoten->type() == type_ListenKnoten) {
    ListenKnoten * lk = (ListenKnoten*) nameKnoten;
    if (lk->getLaenge() == 1) {
      __String varna = lk->getWort(0);
      speicherDefinition(namespace_variable, varna,
          new VarDefinition(varna, wert, vd_konstante)
        );
    }
  }
}



/* Erzeugt eine neue Var-Definition und speichert sie ab. Dabei
   bekommt sie auch gleich eine Nummer. (Aufzurufen, wenn eine
   VarDefinition geparst wurde.) def ist der Default-Wert. */
void DefKnoten::neueVarDefinition(const __String & na, int def) {
  //fprintf(stderr, "Define %s\n", na.data());
  speicherDefinition(namespace_variable, na,
    new VarDefinition(na, def, vd_variable, neueVariable(def)));
}



void DefKnoten::speicherDefinition(int ns, const __String & na,
                                   Definition * f) {
/*  printf("Speichere %s = %s\n", na.data(),
         f->toString().data());*/
  if (mCodeSpeicher[ns].contains(na))
    throw Fehler(_("\"%s\" already defined."), na.data());

  /* (Vielleicht kennt Papi den Code schon, aber das kann man
     berschreiben.) */
     
  mCodeSpeicher[ns][na] = f;
}



/** Liefert eine Code-Definition aus diesem Speicher oder von
    weiter oben. Throwt bei nichtexistenz.
    Achtung: Behlt den Besitz an der Defintion. */
Definition * DefKnoten::getDefinition(int ns, __String na) {

  Definition * ret;


  if (!mCodeSpeicher[ns].contains(na)) {
    /* Wir kennen den Code nicht. Also Papi fragen. */
    if (mVater)
      return mVater->getDefinition(ns, na);
      
    /* Ups, wir haben ja gar keinen Papi. */
    throw Fehler(_("\"%s\" not defined inside << >>."), na.data());
  }
  
  ret = mCodeSpeicher[ns][na];
  
  return ret;
}




/***** Variablen-Nummern-Verwaltung *****/


/** Erzeugt eine unbenannte Variable
    und liefert die Nummer zurck. */
int DefKnoten::neueVariable(int def) {
  /* Sorten-Variablen gibt's nicht (wirklich). Die kommen in den
     Level-Knoten */
  if (mTiefe == tiefe_sorte)
    return mVater->neueVariable(def);

  CASSERT((int) mDefaultWerte.size() == mVarNrBei);
  //fprintf(stderr, "t=%d  %d %d\n", mTiefe, mDefaultWerte.size(), mVarNrBei);
  /* Das kann man auch schneller machen... (immer mehr auf einmal
     reservieren) falls das QT nicht eh tut */
  mDefaultWerte.resize(mVarNrBei + 1);
  mDefaultWerte[mVarNrBei] = def;
  
  return mVarNrBei++;
}




/* Wie war das mit Bitshift und Vorzeichen? */
#define bits_pro_int (8*sizeof(int)-1)


/** Erzeugt eine unbenannte Bool-Variable und liefert
    die Nummer zurck. */
int DefKnoten::neueBoolVariable() {
  /* Sorten-Variablen gibt's nicht (wirklich). Die kommen in den
     Level-Knoten */
  if (mTiefe == tiefe_sorte)
    return mVater->neueBoolVariable();

  if (mBoolNrBei == -1) {
    /* Grad kein Platz mehr fr Bools. Also neuen
       Platz erzeugen. */
    mBoolNrBei = bits_pro_int * neueVariable(0);
  }
  
  int ret = mBoolNrBei;
  mBoolNrBei++;
  
  /* Ist das aktuelle int voll mit Bools? */
  if (mBoolNrBei % bits_pro_int == 0) {
    mBoolNrBei = -1;
  }
  
  return ret;
}



int DefKnoten::getDatenLaenge() const {
  CASSERT(mTiefe == tiefe_level);
  return mVarNrBei;
}


  
/** Liefert den Default-Wert der Variable mit Nummer nr. Es
    muss aber eine richtige Variable sein, die echten Blop-
    Speicherplatz verbraucht. (Sonst soll man sich den Default-
    Wert aus der VarDefinition holen. Das hier ist nur fr
    Variablen-Anfangs-Initialisierung.) */
int DefKnoten::getDefaultWert(int nr) const {
  return mDefaultWerte[nr];
}




/*************************************************************************/
/* ListenKnoten */

__String ListenKnoten::toString() const {
  __String ret;
  const Knoten * k;
  bool mitte = false;
  for (k = mKinder.first(); k != 0; k = mKinder.next()) {
    if (mitte) ret += ", ";
    mitte = true;
   
    ret += k->toString();
  }
  return ret;
}



__String ListenKnoten::getWort(int nr) {
  Knoten * k = mKinder.at(nr);
  CASSERT(k->type() == type_WortKnoten);
  return ((WortKnoten *) k)->getWort();
}



/*************************************************************************/
/* WortKnoten */


__String WortKnoten::toString() const {
  return mWort;
}



